import unittest
from functools import cache

import mpmath
from numpy import inf
from scipy.special import factorial

import osaft
from osaft.core.functions import BesselFunctions as Bf
from osaft.core.functions import conj, full_range, integrate, pi
from osaft.tests.solutions.doinikov1994rigid.setup_test_arf import BaseTestARF


class TestARFGeneral(BaseTestARF):
    def setUp(self) -> None:

        super().setUp()

        self.H_nm_combination_dict = {
            -1: (
                (self.cls.x, self.cls.x_v, ((0, 0),)),
                (self.cls.x_v, self.cls.x, ((0, 0),)),
            ),
            0: (
                (self.cls.x, self.cls.x, ((1, 0), (0, 1))),
                (self.cls.x_v, self.cls.x_v, ((1, 0), (0, 1))),
                (self.cls.x, self.cls.x_v, ((1, 0), (0, 1))),
                (self.cls.x_v, self.cls.x, ((1, 0), (0, 1))),
            ),
            1: (
                (self.cls.x, self.cls.x, ((0, 0),)),
                (self.cls.x, self.cls.x_v, ((0, 0),)),
                (self.cls.x_v, self.cls.x, ((0, 0),)),
            ),
            2: (
                (self.cls.x, self.cls.x, ((1, 0), (0, 1))),
                (self.cls.x_v, self.cls.x_v, ((1, 0), (0, 1))),
                (self.cls.x, self.cls.x_v, ((1, 0), (0, 1))),
                (self.cls.x_v, self.cls.x, ((1, 0), (0, 1))),
            ),
        }

        self.J_nm_combination_dict = {
            -1: ((self.cls.x_v, self.cls.x, ((0, 0),)),),
            0: (
                (self.cls.x, self.cls.x, ((1, 0), (0, 1))),
                (self.cls.x_v, self.cls.x, ((1, 0), (0, 1))),
            ),
            1: (
                (self.cls.x, self.cls.x, ((0, 0),)),
                (self.cls.x_v, self.cls.x, ((0, 0),)),
            ),
            2: (
                (self.cls.x, self.cls.x, ((1, 0), (0, 1))),
                (self.cls.x_v, self.cls.x, ((1, 0), (0, 1))),
            ),
        }

    def test_coefficients(self) -> None:
        # test first for l = 1
        method_dict = {}
        method_dict["G_n_l"] = (1, self.cls.x)  # l = 1
        method_dict["L_n_l"] = 1  # l = 1
        method_dict["K_n_l"] = 1  # l = 1

        self._test_methods_n(method_dict)

        # test first for l = 2
        method_dict = {}
        method_dict["G_n_l"] = (1, self.cls.x)  # l = 2
        method_dict["L_n_l"] = 2  # l = 2
        method_dict["K_n_l"] = 2  # l = 2

        self._test_methods_n(method_dict)

    def test_G_sum_1(self):
        for n in range(0, 5):
            first = self.cls.G_sum_1(n)
            second = self.cls.G_sum_1(n)
            self.assertAlmostEqual(abs(first), abs(second))

    def test_G_sum_2(self):
        for n in range(0, 5):
            first = self.cls.G_sum_2(n)
            second = self.cls.G_sum_2(n)
            self.assertAlmostEqual(abs(first), abs(second))

    def test_G_error(self):
        self.assertRaises(ValueError, self.cls.G_n_l, 1, 3, 1)

    def test_L_sum(self):
        for n in range(0, 5):
            first = self.cls.L_sum(n)
            second = self.cls.L_sum(n)
            self.assertAlmostEqual(abs(first), abs(second))

    def test_L_error(self):
        self.assertRaises(ValueError, self.cls.L_n_l, 1, 3)

    def test_K_sum(self):
        for n in range(0, 5):
            first = self.cls.K_sum(n)
            second = self.cls.K_sum(n)
            self.assertAlmostEqual(abs(first), abs(second))

    def test_K_error(self):
        self.assertRaises(ValueError, self.cls.K_n_l, 1, 3)

    @unittest.skip("Test not working due to numerical problems.")
    def test_G_sum_1_compare(self):
        for n in range(0, 5):
            first = self.cls.G_sum_1(n)
            second = self.cls.G_n_l(n, 1, self.cls.x)
            second += self.cls.G_n_l(n, 2, self.cls.x)
            second /= 2

            self.assertAlmostEqual(abs(first), abs(second), 1e-3)

    @unittest.skip("Test not working due to numerical problems.")
    def test_G_sum_2_compare(self):
        for n in range(0, 4):
            first = self.cls.G_sum_2(n)
            second = self.cls.G_n_l(n, 1, self.cls.x.conjugate()).conjugate()
            second += self.cls.G_n_l(n, 2, self.cls.x)
            second /= 2
            self.assertAlmostEqual(first, second, 1e-5)

    @unittest.skip("Test not working due to numerical problems.")
    def test_L_sum_compare(self):
        for n in range(0, 5):
            first = self.cls.L_sum(n)
            second = self.cls.L_n_l(n, 1)
            second += self.cls.L_n_l(n, 2)
            second /= 2

            self.assertAlmostEqual(abs(first), abs(second), 1e-2)

    @unittest.skip("Test not working due to numerical problems.")
    def test_K_sum_compare(self):
        for n in range(0, 4):
            first = self.cls.K_sum(n)
            second = self.cls.K_n_l(n, 1)
            second += self.cls.K_n_l(n, 2)
            second /= 2
            self.assertAlmostEqual(first, second, 1e-3)

    def test_H_nm_j(self):
        # Testing only combinations of j, (n, m), and x, x_v that are used in
        # the paper
        self.run_integration_test(
            self.cls.H_nm_j,
            self.H_nm_j,
            self.H_nm_combination_dict,
        )

    def test_J_nm_j(self):
        # Testing only combinations of j, (n, m), and x, x_v that are used in
        # the paper
        self.run_integration_test(
            self.cls.J_nm_j,
            self.J_nm_j,
            self.J_nm_combination_dict,
        )

    # -------------------------------------------------------------------------
    # K_n_l
    # -------------------------------------------------------------------------

    def K_n_l(self, n: int, l: int) -> complex:
        if l == 1:
            return self.K_n_1(n)
        else:
            return self.K_n_2(n)

    def K_n_1(self, n: int) -> complex:
        x_con = self.cls.x.conjugate()

        tmp1 = Bf.d1_hankelh1(n, x_con)
        tmp2 = Bf.d1_hankelh1(n + 1, x_con)

        out = (n + 1) * x_con**2 / self.cls.x_v**2 * tmp1
        out *= Bf.hankelh1(n, self.cls.x_v)

        out -= x_con * Bf.d1_hankelh1(n + 1, self.cls.x_v) * tmp2
        return out

    def K_n_2(self, n: int) -> complex:
        x_con = self.cls.x.conjugate()

        tmp1 = Bf.d1_hankelh2(n, x_con)
        tmp2 = Bf.d1_hankelh2(n + 1, x_con)

        out = (n + 1) * x_con**2 / self.cls.x_v**2 * tmp1
        out *= Bf.hankelh1(n, self.cls.x_v)

        out -= x_con * Bf.d1_hankelh1(n + 1, self.cls.x_v) * tmp2
        return out

    def K_sum(self, n: int):
        x_v = self.x_v
        x_con = self.x.conjugate()
        prefactor = (n + 1) * x_con**2 / x_v**2
        term_1 = Bf.hankelh1(n, x_v) * Bf.d1_besselj(n, x_con)
        term_2 = x_con * Bf.d1_hankelh1(n + 1, x_v)
        term_2 *= Bf.d1_besselj(n + 1, x_con)
        return prefactor * term_1 - term_2

    # -------------------------------------------------------------------------
    # L_n_l
    # -------------------------------------------------------------------------

    def L_n_l(self, n: int, l: int) -> complex:
        if l == 1:
            return self.L_n_1(n)
        else:
            return self.L_n_2(n)

    def L_n_1(self, n: int) -> complex:
        x_v_con = self.cls.x_v.conjugate()
        tmp1 = Bf.d1_hankelh1(n, self.cls.x)
        tmp2 = Bf.d1_hankelh1(n + 1, self.cls.x)

        out = -(n + 1) * self.cls.x**2 / self.cls.x_v**2 * tmp2
        out *= Bf.hankelh2(n + 1, x_v_con)

        out += self.cls.x * Bf.d1_hankelh2(n, x_v_con) * tmp1
        return out

    def L_n_2(self, n: int) -> complex:
        x_v_con = self.cls.x_v.conjugate()
        tmp1 = Bf.d1_hankelh2(n, self.cls.x)
        tmp2 = Bf.d1_hankelh2(n + 1, self.cls.x)

        out = -(n + 1) * self.cls.x**2 / self.cls.x_v**2 * tmp2
        out *= Bf.hankelh2(n + 1, x_v_con)

        out += self.cls.x * Bf.d1_hankelh2(n, x_v_con) * tmp1
        return out

    def L_sum(self, n: int) -> complex:
        x = self.x
        x_v = self.x_v
        x_v_con = self.x_v.conjugate()
        term_1 = x * Bf.d1_hankelh2(n, x_v_con) * Bf.d1_besselj(n, x)
        term_2 = (n + 1) * x**2 / x_v**2 * Bf.hankelh2(n + 1, x_v_con)
        term_2 *= Bf.d1_besselj(n + 1, x)
        return term_1 - term_2

    # ------------------------------------------------------------------------
    # G_n_l
    # ------------------------------------------------------------------------

    def G_n_l(self, n: int, l: int, x: complex) -> complex:
        if l == 1:
            return self.G_n_1(n, x)
        else:
            return self.G_n_2(n, x)

    def G_n_1(self, n: int, x: complex) -> complex:
        x_con = x.conjugate()
        tmp1 = Bf.d1_hankelh1(n, x_con)
        tmp2 = Bf.d1_hankelh1(n + 1, x_con)

        out = x_con * Bf.d1_hankelh1(n, x) * tmp1
        out += x * Bf.d1_hankelh1(n + 1, x) * tmp2
        return out

    def G_n_2(self, n: int, x: complex) -> complex:
        x_con = x.conjugate()
        tmp1 = Bf.d1_hankelh2(n, x_con)
        tmp2 = Bf.d1_hankelh2(n + 1, x_con)

        out = x_con * Bf.d1_hankelh1(n, x) * tmp1
        out += x * Bf.d1_hankelh1(n + 1, x) * tmp2
        return out

    def G_sum_1(self, n: int) -> complex:
        x = self.x
        x_con = self.x.conjugate()
        bessel_n = Bf.d1_besselj(n, x_con)
        bessel_n1 = Bf.d1_besselj(n + 1, x_con)
        term_1 = x_con * Bf.d1_hankelh1(n, x) * bessel_n
        term_2 = x * Bf.d1_hankelh1(n + 1, x) * bessel_n1
        return term_1 + term_2

    def G_sum_2(self, n):
        x = self.x
        x_con = self.x.conjugate()
        bessel_n = Bf.d1_besselj(n, x)
        bessel_n1 = Bf.d1_besselj(n + 1, x)
        term_1 = x_con * Bf.d1_hankelh1(n, x_con) * bessel_n
        term_2 = x * Bf.d1_hankelh1(n + 1, x_con) * bessel_n1
        return -(term_1 + term_2)

    # -------------------------------------------------------------------------
    # H_nm_j, J_nm_j, B_qk
    # -------------------------------------------------------------------------

    def H_nm_j(self, x1: complex, x2: complex, n: int, m: int, j: int):

        out = mpmath.mpc(0)
        arg_1 = mpmath.mpc(1j * x2.conjugate() - 1j * x1)
        for k in full_range(n):
            for q in full_range(m):
                M = 2 + k + q + j
                term = mpmath.expint(M, arg_1)
                term *= mpmath.mpc(x1 ** (-(k + 1)))
                term *= mpmath.mpc(x2.conjugate() ** (-(q + 1)))
                term *= self.B_kq(k, q, m, n)
                out += term
        return complex(out)

    @cache
    def J_nm_j(
        self,
        x1: complex,
        x2: complex,
        n: int,
        m: int,
        j: int,
    ) -> complex:
        """Coefficient :math:`J_{nm}^{(j)}`

        :param x1: argument x1
        :param x2: argument x2
        :param n: order n
        :param m: order m
        :param j: exponent
        """
        if 1.01 * abs(x1.real) > abs(x1.imag) > 0.99 * abs(x1.real):
            return self.J_nm_j_quadrature(x1, x2, n, m, j)
        else:
            return self.J_nm_j_sum(x1, x2, n, m, j)

    def J_nm_j_quadrature(
        self,
        x1: complex,
        x2: complex,
        n: int,
        m: int,
        j: int,
    ):
        def kernel(y: float) -> complex:
            out = y ** (-j) * Bf.hankelh1(n, x1 * y)
            out *= Bf.besselj(m, x2.conjugate() * y)
            return out

        return self.integrate_J_nm_H_nm(kernel)

    def J_nm_j_sum(self, x1: complex, x2: complex, n: int, m: int, j: int):

        arg_1 = mpmath.mpc(1j * x2.conjugate() - 1j * x1)
        arg_2 = mpmath.mpc(-1j * x2.conjugate() - 1j * x1)
        out = mpmath.mpc(0)
        for k in full_range(n):
            for q in full_range(m):
                M = 2 + k + q + j
                term = mpmath.expint(M, arg_1)
                term -= (-1) ** (q + m) * mpmath.expint(M, arg_2)
                term *= mpmath.mpc(x1 ** (-(k + 1)))
                term *= mpmath.mpc(x2.conjugate() ** (-(q + 1)))
                term *= self.B_kq(k, q, m, n)
                out += term
        return complex(0.5 * out)

    @staticmethod
    def integrate_J_nm_H_nm(kernel):
        def real_kernel(y: float) -> float:
            return kernel(y).real

        def imag_kernel(y: float) -> float:
            return kernel(y).imag

        real_part, _ = integrate(real_kernel, 1, inf, rel_eps=1e-6)
        imag_part, _ = integrate(imag_kernel, 1, inf, rel_eps=1e-6)

        return real_part + 1j * imag_part

    def run_integration_test(self, fun_1, fun_2, combination_dict):
        range_N_max = full_range(0, 3)
        for j in combination_dict:
            for x1, x2, list_deltas in combination_dict[j]:
                for delta in list_deltas:
                    for n in range_N_max:
                        delta_n, delta_m = delta
                        n, m = n + delta_n, n + delta_m
                        msg = f"{x1 = }, {x2 = }, {j = }, {n = }, {m = }"
                        with self.subTest(msg=msg):
                            self.assertAlmostEqual(
                                fun_1(x1, x2, n, m, j),
                                fun_2(x1, x2, n, m, j),
                                threshold=1e-3,
                            )

    # -------------------------------------------------------------------------
    # B_qk
    # -------------------------------------------------------------------------

    @staticmethod
    def B_kq(k: int, q: int, m: int, n: int) -> complex:
        out = (
            1j ** (m - n)
            * (-1) ** k
            * factorial(n + k)
            * factorial(
                m + q,
            )
        )
        out /= (2 * 1j) ** (k + q) * factorial(k) * factorial(n - k)
        out /= factorial(q) * factorial(m - q)
        return out

    def test_B_kq(self):

        for m in full_range(4):
            for n in full_range(4):
                for q in full_range(m):
                    for k in full_range(n):
                        self.assertEqual(
                            self.cls.B_kq(k, q, m, n),
                            self.B_kq(k, q, m, n),
                        )

    # -------------------------------------------------------------------------
    # S_1n
    # -------------------------------------------------------------------------

    def S_1n(self, n: int) -> complex:

        x = self.cls.x
        x_conj = self.cls.x.conjugate()
        x_v = self.cls.x_v
        H = self.H_nm_j

        first = ((n + 2) * x**2 - n * x_conj**2) / 2
        first *= H(x, x, n, n + 1, 0) - H(x, x, n, n + 1, 2)
        second = x * x_conj * (H(x, x, n + 1, n, 0) - H(x, x, n + 1, n, 2))
        third = x * x_conj / x_v**2 * self.G_n_2(n, x)
        fourth = n * x_conj * H(x, x, n, n, 1)
        fourth += (n + 2) * x * H(x, x, n + 1, n + 1, 1)
        fourth -= x * x_conj * H(x, x, n + 1, n, 0)
        fourth *= (x**2 - x_conj**2) / x_v**2

        return first - second + third + fourth

    def S_2n(self, n: int) -> complex:

        x_v = self.cls.x_v
        x_v_conj = self.cls.x_v.conjugate()
        H = self.H_nm_j

        first = H(x_v, x_v, n + 1, n, 0) - H(x_v, x_v, n + 1, n, 2)
        first *= x_v * x_v_conj
        second = H(x_v, x_v, n, n + 1, 0) - H(x_v, x_v, n, n + 1, 2)
        second *= (n + 1) * x_v**2
        third = Bf.hankelh1(n, x_v) * Bf.d1_hankelh2(n, x_v_conj)
        third += Bf.d1_hankelh1(n + 1, x_v) * Bf.hankelh2(n + 1, x_v_conj)
        third *= n + 1

        return n * (n + 2) * (first - second - third)

    def S_3n(self, n: int) -> complex:

        x = self.cls.x
        x_v = self.cls.x_v
        x_v_conj = self.cls.x_v.conjugate()
        H = self.H_nm_j

        first = H(x, x_v, n, n + 1, 0) - H(x, x_v, n, n + 1, 2)
        first *= 0.5 * (n * x_v_conj**2 - (n + 1) * x**2)

        second = H(x, x_v, n + 1, n, 0) - H(x, x_v, n + 1, n, 2)
        second *= x * x_v_conj

        third = H(x, x_v, n, n, -1) - H(x, x_v, n, n, 1)
        third *= 0.5 * x**2 * x_v_conj

        fourth = H(x, x_v, n + 1, n + 1, -1) - H(x, x_v, n + 1, n + 1, 1)
        fourth *= 0.5 * x * x_v**2

        fifth = n * x_v_conj * H(x, x_v, n, n, 1)
        fifth -= (n + 1) * x * H(x, x_v, n + 1, n + 1, 1)
        fifth *= x**2 / x_v**2

        sixth = self.cls.L_n_l(n, 1)

        return (n + 2) * (first - second + third + fourth + fifth - sixth)

    def S_4n(self, n: int) -> complex:

        x = self.cls.x
        x_conj = self.cls.x.conjugate()
        x_v = self.cls.x_v
        H = self.H_nm_j

        first = H(x_v, x, n, n + 1, 0) - H(x_v, x, n, n + 1, 2)
        first *= ((n + 2) * x_v**2 - (n + 1) * x_conj**2) / 2

        second = H(x_v, x, n + 1, n, 0) - H(x_v, x, n + 1, n, 2)
        second *= x_conj * x_v

        third = H(x_v, x, n, n, -1) - H(x_v, x, n, n, 1)
        third *= 0.5 * x_conj * x_v**2

        fourth = H(x_v, x, n + 1, n + 1, -1) - H(x_v, x, n + 1, n + 1, 1)
        fourth *= x_conj**2 * x_v / 2

        fifth = (n + 1) * x_conj * H(x_v, x, n, n, 1)
        fifth -= (n + 2) * x_v * H(x_v, x, n + 1, n + 1, 1)
        fifth *= x_conj**2 / x_v**2

        sixth = self.cls.K_n_l(n, 2)

        return n * (first + second - third + fourth - fifth + sixth)

    def S_5n(self, n: int) -> complex:

        x = self.cls.x
        x_conj = self.cls.x.conjugate()
        x_v = self.cls.x_v
        J = self.cls.J_nm_j

        first = J(x, x, n, n + 1, 0) - J(x, x, n, n + 1, 2)
        first *= 0.5 * ((n + 2) * x**2 - n * x_conj**2)

        second = J(x, x, n + 1, n, 0) - J(x, x, n + 1, n, 2)
        second *= x * x_conj

        third = self.G_sum_1(n)
        third *= x * x_conj / (x_v**2)

        fourth = n * x_conj * J(x, x, n, n, 1)
        fourth += (n + 2) * x * J(x, x, n + 1, n + 1, 1)
        fourth -= x * x_conj * J(x, x, n + 1, n, 0)
        fourth *= (x**2 - x_conj**2) / x_v**2

        return first - second + third + fourth

    def S_6n(self, n: int) -> complex:

        x = self.cls.x
        x_conj = self.cls.x.conjugate()
        x_v = self.cls.x_v
        J = self.cls.J_nm_j

        first = (J(x, x, n + 1, n, 0) - J(x, x, n + 1, n, 2)).conjugate()
        first *= 0.5 * ((n + 2) * x**2 - n * x_conj**2)

        second = (J(x, x, n, n + 1, 0) - J(x, x, n, n + 1, 2)).conjugate()
        second *= x * x_conj

        third = self.G_sum_2(n)
        third *= x * x_conj / (x_v**2)

        fourth = n * x * J(x, x, n, n, 1)
        fourth += (n + 2) * x_conj * J(x, x, n + 1, n + 1, 1)
        fourth -= x * x_conj * J(x, x, n, n + 1, 0)
        fourth = (x**2 - x_conj**2) / x_v**2 * fourth.conjugate()

        return first - second + third + fourth

    def S_7n(self, n: int) -> complex:

        x = self.cls.x
        x_conj = self.cls.x.conjugate()
        x_v = self.cls.x_v
        J = self.cls.J_nm_j

        first = J(x_v, x, n, n + 1, 0) - J(x_v, x, n, n + 1, 2)
        first *= 0.5 * ((n + 2) * x_v**2 - (n + 1) * x_conj**2)

        second = J(x_v, x, n + 1, n, 0) - J(x_v, x, n + 1, n, 2)
        second *= x_conj * x_v

        third = J(x_v, x, n, n, -1) - J(x_v, x, n, n, 1)
        third *= 0.5 * x_conj * x_v**2

        fourth = J(x_v, x, n + 1, n + 1, -1) - J(x_v, x, n + 1, n + +1, 1)
        fourth *= 0.5 * x_conj**2 * x_v

        fifth = self.K_sum(n)

        sixth = (n + 1) * x_conj * J(x_v, x, n, n, 1)
        sixth -= (n + 2) * x_v * J(x_v, x, n + 1, n + 1, 1)
        sixth *= x_conj**2 / x_v**2

        return n * (first + second - third + fourth + fifth - sixth)

    def S_8n(self, n: int) -> complex:

        x = self.cls.x
        x_conj = self.cls.x.conjugate()
        x_v = self.cls.x_v
        x_v_conj = self.cls.x_v.conjugate()
        J = self.cls.J_nm_j

        first = (J(x_v, x, n + 1, n, 0) - J(x_v, x, n + 1, n, 2)).conjugate()
        first *= 0.5 * (n * x_v_conj**2 - (n + 1) * x**2)

        second = (J(x_v, x, n, n + 1, 0) - J(x_v, x, n, n + 1, 2)).conjugate()
        second *= x * x_v_conj

        third = (J(x_v, x, n, n, -1) - J(x_v, x, n, n, 1)).conjugate()
        third *= 0.5 * x**2 * x_v_conj

        fourth = J(x_v, x, n + 1, n + 1, -1) - J(x_v, x, n + 1, n + 1, 1)
        fourth = x * x_v**2 / 2 * fourth.conjugate()

        fifth = self.L_sum(n)

        sixth = n * x_v * J(x_v, x, n, n, 1)
        sixth -= (n + 1) * x_conj * J(x_v, x, n + 1, n + 1, 1)
        sixth = x**2 / x_v**2 * sixth.conjugate()

        return (n + 2) * (first - second + third + fourth - fifth + sixth)

    def S_9n(self, n: int) -> complex:

        if not self.cls.background_streaming:
            return 0

        x = self.cls.x
        x_conj = self.cls.x.conjugate()
        x_v = self.cls.x_v

        out = x_conj * Bf.d1_besselj(n, x) * Bf.d1_besselj(n, x_conj)
        out += x * Bf.d1_besselj(n + 1, x) * Bf.d1_besselj(n + 1, x_conj)
        out *= x * x_conj / x_v**2

        return out

    def test_S_mn(self):
        names = [f"S_{m}n" for m in range(1, 10)]
        for method in names:
            for n in range(0, 3):
                with self.subTest(method="S_1n", arguments=n):
                    self._test_variables(
                        func_1=self.cls.S_1n,
                        func_2=self.S_1n,
                        args_1=n,
                        args_2=n,
                    )

    # -------------------------------------------------------------------------
    # D_n
    # -------------------------------------------------------------------------

    def D_n(self, n) -> complex:

        a_n = self.cls.alpha_n(n)
        b_n = self.cls.beta_n(n)
        a_n1 = self.cls.alpha_n(n + 1)
        b_n1 = self.cls.beta_n(n + 1)

        D_n = a_n * a_n1.conjugate() * self.cls.S_1n(n)
        D_n += b_n * b_n1.conjugate() * self.cls.S_2n(n)
        D_n -= a_n * b_n1.conjugate() * self.cls.S_3n(n)
        D_n -= b_n * a_n1.conjugate() * self.cls.S_4n(n)
        D_n += a_n * self.cls.S_5n(n)
        D_n += a_n1.conjugate() * self.cls.S_6n(n)
        D_n -= b_n * self.cls.S_7n(n)
        D_n -= b_n1.conjugate() * self.cls.S_8n(n)
        D_n += self.cls.S_9n(n) if self.background_streaming else 0

        return D_n

    def test_D_n(self):

        for n in range(1):
            self.do_testing(self.cls.D_n, self.D_n, n, n)

    # -------------------------------------------------------------------------
    # Test ARF
    # -------------------------------------------------------------------------

    def compute_arf(self):
        def term(n: int):
            out = self.D_n(n) * self.cls.A_in(n) * conj(self.cls.A_in(n + 1))

            tmp = conj(self.D_n(n)) * self.cls.A_in(n + 1)
            tmp *= conj(self.cls.A_in(n))

            out += tmp
            out *= (n + 1) / (2 * n + 1) / (2 * n + 3)
            return out

        F_1 = sum([term(n) for n in full_range(self.N_max)]).real
        if self.cls.wave_type == osaft.WaveType.STANDING:
            F_2 = self.cls._F_2_standing()
        else:
            F_2 = self.cls._F_2_travelling()
        return -3 / 2 * pi * self.rho_f * F_1 + F_2

    def test_wrong_combination_error(self):
        self.cls.small_boundary_layer = True
        self.cls.large_boundary_layer = True
        self.assertRaises(ValueError, self.cls.compute_arf)

    def test_arf(self):
        self.do_testing(
            self.cls.compute_arf,
            self.compute_arf,
            threshold=1e-10,
        )


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import numpy as np

from osaft.core.backgroundfields import WaveType
from osaft.core.frequency import Frequency
from osaft.core.functions import BesselFunctions as Bf
from osaft.core.geometries import Sphere
from osaft.core.variable import ActiveListVariable, ActiveVariable
from osaft.solutions.doinikov1994compressible.base import (
    BaseDoinikov1994Compressible,
)

NDArray = np.ndarray


class CoefficientMatrix(BaseDoinikov1994Compressible):
    """Coefficient Matrix Doinikov (viscous fluid-viscous sphere; 1994)

    :param f: Frequency [Hz]
    :param R_0: Radius of the sphere [m]
    :param rho_s: Density of the sphere [kg/m^3]
    :param c_s: Speed of sound of in the sphere [m/s]
    :param eta_s: shear viscosity of in the sphere  [Pa s]
    :param zeta_s: bulk viscosity of in the sphere [Pa s]
    :param rho_f: Density of the fluid [kg/m^3]
    :param c_f: Speed of sound of the fluid [m/s]
    :param eta_f: shear viscosity [Pa s]
    :param zeta_f: bulk viscosity [Pa s]
    :param p_0: Pressure amplitude of the field [Pa]
    :param position: Position within the standing wave field [m]
    :param wave_type: Type of wave, traveling or standing
    """

    def __init__(
        self,
        f: Frequency | float | int,
        R_0: Sphere | float | int,
        rho_s: float,
        c_s: float,
        eta_s: float,
        zeta_s: float,
        rho_f: float,
        c_f: float,
        eta_f: float,
        zeta_f: float,
        p_0: float,
        wave_type: WaveType,
        position: float,
    ) -> None:

        # init of parent method
        super().__init__(
            f=f,
            R_0=R_0,
            rho_s=rho_s,
            c_s=c_s,
            eta_s=eta_s,
            zeta_s=zeta_s,
            rho_f=rho_f,
            c_f=c_f,
            eta_f=eta_f,
            zeta_f=zeta_f,
            p_0=p_0,
            wave_type=wave_type,
            position=position,
        )

        # Dependent Variables
        self._x_hat = ActiveVariable(self._compute_x_hat, "x_hat")
        self._x_hat_v = ActiveVariable(self._compute_x_hat_v, "x_v hat")
        self._list_matrix_M_n = ActiveListVariable(
            self._compute_matrix_M_n,
            "Matrices M(n)",
        )
        self._list_vector_N_n = ActiveListVariable(
            self._compute_vector_N_n,
            "Vectors N(n)",
        )
        self._x_hat.is_computed_by(
            self.scatterer._k_f,
            self.sphere._R_0,
        )
        self._x_hat_v.is_computed_by(
            self.scatterer._k_v,
            self.sphere._R_0,
        )
        self._list_matrix_M_n.is_computed_by(
            self._x,
            self._x_v,
            self._x_hat,
            self._x_hat_v,
            self.frequency._omega,
            self.fluid._c_f,
            self.scatterer._c_f,
            self.fluid._rho_f,
            self.scatterer._rho_f,
            self.fluid._eta_f,
            self.scatterer._eta_f,
            self.fluid._zeta_f,
            self.scatterer._zeta_f,
        )
        self._list_vector_N_n.is_computed_by(
            self._x,
            self.frequency._omega,
            self.fluid._c_f,
            self.fluid._rho_f,
            self.fluid._eta_f,
            self.fluid._zeta_f,
        )

    def det_M_n(self, n: int, column: None | int = None) -> complex:
        """Determinant of the matrix `M` for the mode `n`

        :param n: mode
        :param column: the `l`th coefficient is replaced with the vector `N`
        """
        matrix = self.M(n).copy()
        vector = self.N(n)
        if column is not None:
            matrix[:, column] = vector
        return np.linalg.det(matrix)

    def M(self, n: int) -> NDArray:
        """Matrix M of order `n`

        :param n: order
        """
        return self._list_matrix_M_n.item(n)

    def N(self, n: int) -> NDArray:
        """Vector `N`"""
        return self._list_vector_N_n.item(n)

    # -------------------------------------------------------------------------
    # Properties
    # -------------------------------------------------------------------------

    @property
    def x_hat(self) -> complex:
        """Product of :attr:`~.k_s` and :attr:`~.R_0`
        :math:`\\hat{x}=k_s R_0`
        """
        return self._x_hat.value

    def _compute_x_hat(self) -> complex:
        return self.k_s * self.R_0

    @property
    def x_hat_v(self) -> complex:
        """Product of :attr:`~.k_vs` and :attr:`~.R_0`
        :math:`\\hat{x}_v=\\hat{k}_s R_0`
        """
        return self._x_hat_v.value

    def _compute_x_hat_v(self) -> complex:
        return self.k_vs * self.R_0

    # -------------------------------------------------------------------------
    # Vector N(n)
    # -------------------------------------------------------------------------

    def _compute_vector_N_n(self, n: int) -> NDArray:
        vector = np.zeros(4, dtype=complex)

        vector[0] = -self.x * Bf.d1_besselj(n, self.x)

        vector[1] = -Bf.besselj(n, self.x)
        vector[2] = 1j * self.rho_f * self.c_f**2 / self.omega
        vector[2] += self.zeta_f
        vector[2] -= 2 / 3 * self.eta_f
        vector[2] *= Bf.besselj(n, self.x)
        vector[2] -= 2 * self.eta_f * Bf.d2_besselj(n, self.x)
        vector[2] *= self.x**2
        vector[3] = Bf.besselj(n, self.x)
        vector[3] -= self.x * Bf.d1_besselj(n, self.x)
        vector[3] *= 2 * self.eta_f

        return vector

    # -----------------------------------------------------------------------
    # Matrix M(n)
    # -----------------------------------------------------------------------

    def _compute_matrix_M_n(self, n: int) -> NDArray:

        # Initialize matrix
        matrix = np.zeros((4, 4), dtype=complex)

        # First row
        matrix[0, 0] = self.x * Bf.d1_hankelh1(n, self.x)
        matrix[0, 1] = -n * (n + 1) * Bf.hankelh1(n, self.x_v)
        matrix[0, 2] = -self.x_hat * Bf.d1_besselj(n, self.x_hat)
        matrix[0, 3] = n * (n + 1) * Bf.besselj(n, self.x_hat_v)

        # Second Row
        matrix[1, 0] = Bf.hankelh1(n, self.x)
        matrix[1, 1] = -Bf.hankelh1(n, self.x_v)
        matrix[1, 1] -= self.x_v * Bf.d1_hankelh1(n, self.x_v)
        matrix[1, 2] = -Bf.besselj(n, self.x_hat)
        matrix[1, 3] = Bf.besselj(n, self.x_hat_v)
        matrix[1, 3] += self.x_hat_v * Bf.d1_besselj(n, self.x_hat_v)

        # Third row
        matrix[2, 0] = 1j * self.rho_f * self.c_f**2 / self.omega
        matrix[2, 0] += self.zeta_f
        matrix[2, 0] -= 2 / 3 * self.eta_f
        matrix[2, 0] *= -Bf.hankelh1(n, self.x)
        matrix[2, 0] += 2 * self.eta_f * Bf.d2_hankelh1(n, self.x)
        matrix[2, 0] *= self.x**2

        matrix[2, 1] = Bf.hankelh1(n, self.x_v)
        matrix[2, 1] -= self.x_v * Bf.d1_hankelh1(n, self.x_v)
        matrix[2, 1] *= 2 * n * (n + 1) * self.eta_f

        matrix[2, 2] = 1j * self.rho_s * self.c_s**2 / self.omega
        matrix[2, 2] += self.zeta_s
        matrix[2, 2] -= 2 / 3 * self.eta_s
        matrix[2, 2] *= Bf.besselj(n, self.x_hat)
        matrix[2, 2] -= 2 * self.eta_s * Bf.d2_besselj(n, self.x_hat)
        matrix[2, 2] *= self.x_hat**2

        matrix[2, 3] = self.x_hat_v * Bf.d1_besselj(n, self.x_hat_v)
        matrix[2, 3] -= Bf.besselj(n, self.x_hat_v)
        matrix[2, 3] *= 2 * n * (n + 1) * self.eta_s

        # Fourth row
        matrix[3, 0] = self.x * Bf.d1_hankelh1(n, self.x)
        matrix[3, 0] -= Bf.hankelh1(n, self.x)
        matrix[3, 0] *= 2 * self.eta_f
        matrix[3, 1] = self.x_v**2 * Bf.d2_hankelh1(n, self.x_v)
        matrix[3, 1] += (n**2 + n - 2) * Bf.hankelh1(n, self.x_v)
        matrix[3, 1] *= -self.eta_f
        matrix[3, 2] = Bf.besselj(n, self.x_hat)
        matrix[3, 2] -= self.x_hat * Bf.d1_besselj(n, self.x_hat)
        matrix[3, 2] *= 2 * self.eta_s

        matrix[3, 3] = self.x_hat_v**2 * Bf.d2_besselj(n, self.x_hat_v)
        matrix[3, 3] += (n**2 + n - 2) * Bf.besselj(n, self.x_hat_v)
        matrix[3, 3] *= self.eta_s
        return matrix


if __name__ == "__main__":
    pass



import warnings

from osaft import log

EPS = 1e-2


class AssumptionWarning(Warning):
    """Warning that the used solution might not be valid based on the set
    material parameters
    """

    pass


def raise_assumption_warning(test: bool, msg: str = "") -> None:
    """Raises :class:`AssumptionWarning` if :attr:`test` is `True`

    :param test: test that checks for warning
    :param msg: message for more descriptive warning message
    """

    if test:
        message = "Theory might not be valid anymore!"
        message += msg

        warnings.warn(message, category=AssumptionWarning, stacklevel=5)
        log.info(message)


if __name__ == "__main__":
    pass



import unittest

import numpy as np
from matplotlib import pyplot as plt

from osaft.plotting.grid import (
    CartesianFluidGrid,
    CartesianGrid,
    CartesianScattererGrid,
)
from osaft.tests.plotting.grid.test_base_grid import plot_grid


class TestCartesianGrid(unittest.TestCase):
    def setUp(self) -> None:
        self.r_max = 2
        self.res_1 = 10
        self.res_2 = 20

        # Grid with the same resolution in both dimensions
        self.grid_1_res = CartesianGrid(
            self.r_max,
            self.res_1,
            False,
            False,
        )

        # Grid with the different resolutions in the two dimensions
        self.grid_2_res = CartesianGrid(
            self.r_max,
            (self.res_1, self.res_2),
            False,
            False,
        )

        # Grid with the same resolution in both dimensions
        self.grid_1_res_upper = CartesianGrid(
            self.r_max,
            self.res_1,
            True,
            False,
        )
        # Grid with the different resolutions in the two dimensions
        self.grid_2_res_upper = CartesianGrid(
            self.r_max,
            (self.res_1, self.res_2),
            True,
            False,
        )

    # -------------------------------------------------------------------------
    # Visual tests (skipped)
    # -------------------------------------------------------------------------

    @unittest.skip("Used only for visual testing during development.")
    def test_plot(self):
        plot_grid(self.grid_1_res)
        plot_grid(self.grid_2_res_upper)
        plot_grid(self.grid_1_res)
        plot_grid(self.grid_2_res_upper)
        plt.show()

    # -------------------------------------------------------------------------
    # Test step size tests
    # -------------------------------------------------------------------------

    def test_dz(self):
        dz = 2 * self.r_max / (self.res_1 - 1)
        self.assertEqual(self.grid_1_res.dz, dz)
        self.assertEqual(self.grid_2_res.dz, dz)

    def test_dx(self):
        dx_1 = 2 * self.r_max / (self.res_1 - 1)
        dx_2 = 2 * self.r_max / (self.res_2 - 1)
        self.assertEqual(self.grid_1_res.dx, dx_1)
        self.assertEqual(self.grid_2_res.dx, dx_2)

        dx_1 = self.r_max / (self.res_1 // 2 - 1)
        dx_2 = self.r_max / (self.res_2 - 1)
        self.assertEqual(self.grid_1_res_upper.dx, dx_1)
        self.assertEqual(self.grid_2_res_upper.dx, dx_2)

    # -------------------------------------------------------------------------
    # Array / resolution tests
    # -------------------------------------------------------------------------

    def test_arr_z(self) -> None:
        arr_z = np.linspace(-self.r_max, self.r_max, self.res_1)
        np.testing.assert_array_equal(self.grid_1_res.arr_z, arr_z)
        np.testing.assert_array_equal(self.grid_2_res.arr_z, arr_z)

    def test_arr_x(self) -> None:
        arr_x_1 = np.linspace(-self.r_max, self.r_max, self.res_1)
        arr_x_2 = np.linspace(-self.r_max, self.r_max, self.res_2)
        np.testing.assert_array_equal(self.grid_1_res.arr_x, arr_x_1)
        np.testing.assert_array_equal(self.grid_2_res.arr_x, arr_x_2)

        arr_x_1 = np.linspace(0, self.r_max, self.res_1 // 2)
        arr_x_2 = np.linspace(0, self.r_max, self.res_2)
        np.testing.assert_array_equal(
            self.grid_1_res_upper.arr_x,
            arr_x_1,
        )
        np.testing.assert_array_equal(
            self.grid_2_res_upper.arr_x,
            arr_x_2,
        )

    # -------------------------------------------------------------------------
    # Test offset
    # -------------------------------------------------------------------------

    def test_offset(self) -> None:
        self.grid_with_offset = CartesianGrid(
            self.r_max,
            5,
            True,
            True,
        )
        self.assertEqual(np.abs(self.grid_with_offset.arr_z).min(), 1e-30)

        self.grid_without_offset = CartesianGrid(
            self.r_max,
            5,
            True,
            False,
        )
        self.assertEqual(np.abs(self.grid_without_offset.arr_z).min(), 0)

    # -------------------------------------------------------------------------
    # Test meshes
    # -------------------------------------------------------------------------

    def test_meshes(self) -> None:

        # Upper = False
        arr_x = self.grid_1_res.arr_x
        arr_z = self.grid_1_res.arr_z
        mesh_z, mesh_x = np.meshgrid(arr_z, arr_x)

        np.testing.assert_array_equal(mesh_x, self.grid_1_res.mesh_x)
        np.testing.assert_array_equal(mesh_z, self.grid_1_res.mesh_z)

        # Upper = True
        arr_x = self.grid_1_res_upper.arr_x
        arr_z = self.grid_1_res_upper.arr_z
        mesh_z, mesh_x = np.meshgrid(arr_z, arr_x)

        np.testing.assert_array_equal(mesh_x, self.grid_1_res_upper.mesh_x)
        np.testing.assert_array_equal(mesh_z, self.grid_1_res_upper.mesh_z)

    # -------------------------------------------------------------------------
    # Test conversions
    # -------------------------------------------------------------------------

    def test_conversion(self) -> None:

        # Upper = False
        mesh_x = self.grid_1_res.mesh_x
        mesh_z = self.grid_1_res.mesh_z

        mesh_r = np.hypot(mesh_x, mesh_z)
        mesh_t = np.arctan2(mesh_x, mesh_z)

        np.testing.assert_array_equal(mesh_r, self.grid_1_res.mesh_r)
        np.testing.assert_array_equal(mesh_t, self.grid_1_res.mesh_theta)

        # Upper = True
        mesh_x = self.grid_2_res_upper.mesh_x
        mesh_z = self.grid_2_res_upper.mesh_z

        mesh_r = np.hypot(mesh_x, mesh_z)
        mesh_t = np.arctan2(mesh_x, mesh_z)

        np.testing.assert_array_equal(mesh_r, self.grid_2_res_upper.mesh_r)
        np.testing.assert_array_equal(mesh_t, self.grid_2_res_upper.mesh_theta)


class TestFluidCartesianGrid(unittest.TestCase):
    def setUp(self) -> None:
        self.r_min = 1
        self.r_max = 2
        self.res_1 = 20
        self.res_2 = 40

        # Grid with the same resolution in both dimensions
        self.grid_1_res = CartesianFluidGrid(
            self.r_min,
            self.r_max,
            self.res_1,
            False,
        )

        # Grid with the different resolutions in the two dimensions
        self.grid_2_res = CartesianFluidGrid(
            self.r_min,
            self.r_max,
            (self.res_1, self.res_2),
            False,
        )

        # Grid with the same resolution in both dimensions
        self.grid_1_res_upper = CartesianFluidGrid(
            self.r_min,
            self.r_max,
            self.res_1,
            True,
        )
        # Grid with the different resolutions in the two dimensions
        self.grid_2_res_upper = CartesianFluidGrid(
            self.r_min,
            self.r_max,
            (self.res_1, self.res_2),
            True,
        )

    @unittest.skip("Used only for visual testing during development.")
    def test_plot(self):
        plot_grid(self.grid_1_res)
        plot_grid(self.grid_2_res_upper)
        plot_grid(self.grid_1_res)
        plot_grid(self.grid_2_res_upper)
        plt.show()

    def test_mask(self):

        self.assertTrue((self.grid_1_res.mesh_r >= self.r_min).all())
        self.assertTrue((self.grid_1_res_upper.mesh_r >= self.r_min).all())
        self.assertTrue((self.grid_2_res.mesh_r >= self.r_min).all())
        self.assertTrue((self.grid_2_res_upper.mesh_r >= self.r_min).all())

        self.assertTrue(
            (
                np.hypot(
                    self.grid_1_res.mesh_x,
                    self.grid_1_res.mesh_z,
                )
                >= self.r_min
            ).all(),
        )
        self.assertTrue(
            (
                np.hypot(
                    self.grid_1_res_upper.mesh_x,
                    self.grid_1_res_upper.mesh_z,
                )
                >= self.r_min
            ).all(),
        )
        self.assertTrue(
            (
                np.hypot(
                    self.grid_2_res.mesh_x,
                    self.grid_2_res.mesh_z,
                )
                >= self.r_min
            ).all(),
        )
        self.assertTrue(
            (
                np.hypot(
                    self.grid_2_res_upper.mesh_x,
                    self.grid_2_res_upper.mesh_z,
                )
                >= self.r_min
            ).all(),
        )


class TestScattererCartesianGrid(unittest.TestCase):
    def setUp(self) -> None:
        self.r_min = 1
        self.r_max = 2
        self.res_1 = 20
        self.res_2 = 40

        # Grid with the same resolution in both dimensions
        self.grid_1_res = CartesianScattererGrid(
            self.r_max,
            self.res_1,
            False,
        )
        # Grid with the different resolutions in the two dimensions
        self.grid_2_res = CartesianScattererGrid(
            self.r_max,
            (self.res_1, self.res_2),
            False,
        )
        # Grid with the same resolution in both dimensions
        self.grid_1_res_upper = CartesianScattererGrid(
            self.r_max,
            self.res_1,
            True,
        )
        # Grid with the different resolutions in the two dimensions
        self.grid_2_res_upper = CartesianScattererGrid(
            self.r_max,
            (self.res_1, self.res_2),
            True,
        )

    @unittest.skip("Used only for visual testing during development.")
    def test_plot(self):
        plot_grid(self.grid_1_res)
        plot_grid(self.grid_2_res_upper)
        plot_grid(self.grid_1_res)
        plot_grid(self.grid_2_res_upper)
        plt.show()

    def test_mask(self):

        self.assertTrue((self.grid_1_res.mesh_r <= self.r_max).all())
        self.assertTrue((self.grid_1_res_upper.mesh_r <= self.r_max).all())
        self.assertTrue((self.grid_2_res.mesh_r <= self.r_max).all())
        self.assertTrue((self.grid_2_res_upper.mesh_r <= self.r_max).all())

        self.assertTrue(
            (
                np.hypot(
                    self.grid_1_res.mesh_x,
                    self.grid_1_res.mesh_z,
                )
                <= self.r_max
            ).all(),
        )
        self.assertTrue(
            (
                np.hypot(
                    self.grid_1_res_upper.mesh_x,
                    self.grid_1_res_upper.mesh_z,
                )
                <= self.r_max
            ).all(),
        )
        self.assertTrue(
            (
                np.hypot(
                    self.grid_2_res.mesh_x,
                    self.grid_2_res.mesh_z,
                )
                <= self.r_max
            ).all(),
        )
        self.assertTrue(
            (
                np.hypot(
                    self.grid_2_res_upper.mesh_x,
                    self.grid_2_res_upper.mesh_z,
                )
                <= self.r_max
            ).all(),
        )


if __name__ == "__main__":
    unittest.main()



import unittest

from osaft.core.functions import BesselFunctions as Bf
from osaft.tests.basetest_numeric import BaseTestSolutions
from osaft.tests.solution_factory import SolutionFactory


class TestCoefficientMatrix(BaseTestSolutions):
    def setUp(self) -> None:
        super().setUp()

        self._v_particle_threshold = 1e-5
        self._v_fluid_threshold = 1e-3
        self._v_boundary_conditions = 1e-10

        self.cls = SolutionFactory().doinikov_1994_compressible_scattering()

    @property
    def x(self):
        return self.cls.x

    @property
    def x_v(self):
        return self.cls.x_v

    @property
    def x_hat(self):
        return self.cls.x_hat

    @property
    def x_hat_v(self):
        return self.cls.x_hat_v

    def test_matrix_coefficients(self) -> None:

        for k in range(1, 5):
            for l in range(1, 5):
                for n in range(3):
                    with self.subTest(msg=f"M_{k}{l}", n=n):
                        func_1 = getattr(self, f"m_{k}{l}")
                        func_2 = getattr(self, f"cls_m_{k}{l}")
                        self.do_testing(func_1, func_2, n, n, 1e-6)

    def test_vector_coefficients(self):
        for k in range(1, 5):
            for n in range(3):
                with self.subTest(msg=f"N_{k}", n=n):
                    func_1 = getattr(self, f"n_{k}")
                    func_2 = getattr(self, f"cls_n_{k}")
                    self.do_testing(func_1, func_2, n, n, 1e-6)

    # ------------------
    # testing m_1i
    # ------------------

    def cls_m_11(self, n: int) -> complex:
        return self.cls.M(n)[0, 0]

    def m_11(self, n: int) -> complex:
        out = self.x * Bf.d1_hankelh1(n, self.x)
        return out

    def cls_m_12(self, n: int) -> complex:
        return self.cls.M(n)[0, 1]

    def m_12(self, n: int) -> complex:
        out = -n * (n + 1) * Bf.hankelh1(n, self.x_v)
        return out

    def cls_m_13(self, n: int) -> complex:
        return self.cls.M(n)[0, 2]

    def m_13(self, n: int) -> complex:
        out = -self.x_hat * Bf.d1_besselj(n, self.x_hat)
        return out

    def cls_m_14(self, n: int) -> complex:
        return self.cls.M(n)[0, 3]

    def m_14(self, n: int) -> complex:
        out = n * (n + 1) * Bf.besselj(n, self.x_hat_v)
        return out

    # ------------------
    # testing m_2i
    # ------------------

    def cls_m_21(self, n: int) -> complex:
        return self.cls.M(n)[1, 0]

    def m_21(self, n: int) -> complex:
        out = Bf.hankelh1(n, self.x)
        return out

    def cls_m_22(self, n: int) -> complex:
        return self.cls.M(n)[1, 1]

    def m_22(self, n: int) -> complex:
        out = -Bf.hankelh1(n, self.x_v)
        out -= self.x_v * Bf.d1_hankelh1(n, self.x_v)
        return out

    def cls_m_23(self, n: int) -> complex:
        return self.cls.M(n)[1, 2]

    def m_23(self, n: int) -> complex:
        out = -Bf.besselj(n, self.x_hat)
        return out

    def cls_m_24(self, n: int) -> complex:
        return self.cls.M(n)[1, 3]

    def m_24(self, n: int) -> complex:
        out = Bf.besselj(n, self.x_hat_v)
        out += self.x_hat_v * Bf.d1_besselj(n, self.x_hat_v)
        return out

    # ------------------
    # testing m_3i
    # ------------------

    def cls_m_31(self, n: int) -> complex:
        return self.cls.M(n)[2, 0]

    def m_31(self, n: int) -> complex:
        out = 1j * self.cls.rho_f * self.cls.c_f**2 / self.cls.omega
        out += self.cls.zeta_f
        out -= 2 * self.cls.eta_f / 3
        out *= -Bf.hankelh1(n, self.x)

        out += 2 * self.cls.eta_f * Bf.d2_hankelh1(n, self.x)

        out *= self.x**2

        return out

    def cls_m_32(self, n: int) -> complex:
        return self.cls.M(n)[2, 1]

    def m_32(self, n: int) -> complex:
        out = Bf.hankelh1(n, self.x_v)
        out -= self.x_v * Bf.d1_hankelh1(n, self.x_v)

        out *= 2 * n * (n + 1) * self.cls.eta_f
        return out

    def cls_m_33(self, n: int) -> complex:
        return self.cls.M(n)[2, 2]

    def m_33(self, n: int) -> complex:
        out = 1j * self.cls.rho_s * self.cls.c_s**2 / self.cls.omega
        out += self.cls.zeta_s
        out -= 2 * self.cls.eta_s / 3
        out *= Bf.besselj(n, self.x_hat)

        out -= 2 * self.cls.eta_s * Bf.d2_besselj(n, self.x_hat)

        out *= self.x_hat**2
        return out

    def cls_m_34(self, n: int) -> complex:
        return self.cls.M(n)[2, 3]

    def m_34(self, n: int) -> complex:
        out = -Bf.besselj(n, self.x_hat_v)
        out += self.x_hat_v * Bf.d1_besselj(n, self.x_hat_v)

        out *= 2 * n * (n + 1) * self.cls.eta_s
        return out

    # ------------------
    # testing m_4i
    # ------------------

    def cls_m_41(self, n: int) -> complex:
        return self.cls.M(n)[3, 0]

    def m_41(self, n: int) -> complex:
        out = self.x * Bf.d1_hankelh1(n, self.x)
        out -= Bf.hankelh1(n, self.x)

        out *= 2 * self.cls.eta_f

        return out

    def cls_m_42(self, n: int) -> complex:
        return self.cls.M(n)[3, 1]

    def m_42(self, n: int) -> complex:
        out = (n**2 + n - 2) * Bf.hankelh1(n, self.x_v)

        out += self.x_v**2 * Bf.d2_hankelh1(n, self.x_v)

        out *= -self.cls.eta_f
        return out

    def cls_m_43(self, n: int) -> complex:
        return self.cls.M(n)[3, 2]

    def m_43(self, n: int) -> complex:
        out = Bf.besselj(n, self.x_hat)
        out -= self.x_hat * Bf.d1_besselj(n, self.x_hat)

        out *= 2 * self.cls.eta_s
        return out

    def cls_m_44(self, n: int) -> complex:
        return self.cls.M(n)[3, 3]

    def m_44(self, n: int) -> complex:
        out = (n**2 + n - 2) * Bf.besselj(n, self.x_hat_v)

        out += self.x_hat_v**2 * Bf.d2_besselj(n, self.x_hat_v)

        out *= self.cls.eta_s
        return out

    # ------------------
    # testing n_i
    # ------------------

    def cls_n_1(self, n: int) -> complex:
        return self.cls.N(n)[0]

    def n_1(self, n: int) -> complex:
        out = -self.x * Bf.d1_besselj(n, self.x)
        return out

    def cls_n_2(self, n: int) -> complex:
        return self.cls.N(n)[1]

    def n_2(self, n: int) -> complex:
        out = -Bf.besselj(n, self.x)
        return out

    def cls_n_3(self, n: int) -> complex:
        return self.cls.N(n)[2]

    def n_3(self, n: int) -> complex:
        out = 1j * self.cls.rho_f * self.cls.c_f**2 / self.cls.omega
        out += self.cls.zeta_f
        out -= 2 * self.cls.eta_f / 3
        out *= Bf.besselj(n, self.x)
        out -= 2 * self.cls.eta_f * Bf.d2_besselj(n, self.x)
        out *= self.x**2

        return out

    def cls_n_4(self, n: int) -> complex:
        return self.cls.N(n)[3]

    def n_4(self, n: int) -> complex:
        out = Bf.besselj(n, self.x)
        out -= self.x * Bf.d1_besselj(n, self.x)
        out *= 2 * self.cls.eta_f
        return out

    if __name__ == "__main__":
        unittest.main()



import unittest

from osaft import WaveType, doinikov1994compressible, yosioka1955
from osaft.core.basecomposite import (
    BaseFrequencyComposite,
    BaseSphereFrequencyComposite,
)
from osaft.core.frequency import Frequency
from osaft.core.functions import pi
from osaft.core.geometries import Sphere
from osaft.core.solids import ElasticSolid
from osaft.tests.basetest_numeric import BaseTestSolutions


class TestBaseFrequencyComposite(BaseTestSolutions):
    def test_types_for_frequency(self):
        cls = BaseFrequencyComposite(10.1)
        self.assertTrue(isinstance(cls, BaseFrequencyComposite))
        cls = BaseFrequencyComposite(10.1)
        self.assertTrue(isinstance(cls, BaseFrequencyComposite))

        F = Frequency(1e6)
        cls = BaseFrequencyComposite(F)
        self.assertTrue(isinstance(cls, BaseFrequencyComposite))

    def test_TypeError(self):

        self.assertRaises(TypeError, BaseFrequencyComposite, "s")

    def test_input_variables(self):

        # Test elastic solid
        solid = ElasticSolid(1e6, 1e6, 0.3, 1e3)
        expected_solid = {"E_s", "nu_s", "rho_s", "f"}
        got_solid = set(solid.input_variables())
        self.assertSetEqual(got_solid, expected_solid)

        # Doinikov Compressible
        doinikov = doinikov1994compressible.ARF(
            1e6,
            1e-6,
            3e3,
            1e3,
            1e-3,
            1e-3,
            1e3,
            1e3,
            1e-3,
            1e-3,
            1e5,
            WaveType.STANDING,
            0,
            True,
        )

        # Test Doinikov 1994 Compressible
        expected_doinikov = {
            "small_boundary_layer",
            "large_boundary_layer",
            "rho_s",
            "c_s",
            "eta_s",
            "zeta_s",
            "position",
            "p_0",
            "wave_type",
            "rho_f",
            "c_f",
            "eta_f",
            "zeta_f",
            "R_0",
            "f",
            "N_max",
            "background_streaming",
        }
        got_doinikov = set(doinikov.input_variables())
        self.assertSetEqual(got_doinikov, expected_doinikov)

        # Test Yosioka & Kawasima (1955)
        yosioka = yosioka1955.ARF(
            1e6,
            1e-6,
            2e3,
            2e3,
            1e3,
            1.5e3,
            1e5,
            WaveType.TRAVELLING,
        )
        expected_yosioka = {
            "small_particle",
            "bubble_solution",
            "position",
            "p_0",
            "wave_type",
            "rho_s",
            "c_s",
            "rho_f",
            "c_f",
            "R_0",
            "f",
            "N_max",
        }
        got_yosioka = set(yosioka.input_variables())
        self.assertSetEqual(got_yosioka, expected_yosioka)


class TestBaseSphereFrequencyComposite(BaseTestSolutions):
    def test_types_for_sphere(self):
        cls = BaseSphereFrequencyComposite(10.1, 10)
        self.assertTrue(isinstance(cls, BaseSphereFrequencyComposite))
        cls = BaseSphereFrequencyComposite(10.1, 10.0)
        self.assertTrue(isinstance(cls, BaseSphereFrequencyComposite))

        S = Sphere(1e-6)
        cls = BaseSphereFrequencyComposite(10.1, S)
        self.assertTrue(isinstance(cls, BaseSphereFrequencyComposite))

    def test_properties(self):
        def V(R):
            return 4 / 3 * R**3 * pi

        def A(R):
            return 4 * R**2 * pi

        R = 1e-6
        f = 1e6
        cls = BaseSphereFrequencyComposite(f, R)

        self.assertAlmostEqual(R, cls.R_0)
        self.assertAlmostEqual(V(R), cls.volume)
        self.assertAlmostEqual(A(R), cls.area)

        R = 4.39e-6
        cls.R_0 = R
        self.assertAlmostEqual(R, cls.R_0)
        self.assertAlmostEqual(V(R), cls.volume)
        self.assertAlmostEqual(A(R), cls.area)

    def test_TypeError(self):
        self.assertRaises(
            TypeError,
            BaseSphereFrequencyComposite,
            10.0,
            "s",
        )


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
from scipy.special import hankel1, hankel2, jv, yv

from osaft.core.functions import BesselFunctions as Bf
from osaft.core.functions import pi, sqrt
from osaft.tests.basetest_numeric import NumericTestCase


class TestBesselFunctions(NumericTestCase):
    def setUp(self) -> None:

        self.rng = np.random.default_rng()

    def random_z(self, low: float = -10, high: float = 10) -> complex:
        a = self.rng.uniform(low, high)
        b = self.rng.uniform(low, high)
        return a + b * 1j

    def test_besselj_and_derivatives(self):
        def j(n, z):
            return sqrt(pi / 2 / z) * jv(n + 0.5, z)

        for n in self.rng.integers(0, 10, size=10):
            for _ in np.arange(10):
                z = self.random_z()
                self.assertAlmostEqual(j(n, z), Bf.besselj(n, z), 1e-9)
                self.assertAlmostEqual(
                    self._df(j, n, z),
                    Bf.d1_besselj(n, z),
                    1e-9,
                )
                self.assertAlmostEqual(
                    self._d2f(j, n, z),
                    Bf.d2_besselj(n, z),
                    1e-9,
                )
                self.assertAlmostEqual(
                    self._d3f(j, n, z),
                    Bf.d3_besselj(n, z),
                    1e-9,
                )
                self.assertAlmostEqual(
                    self._d4f(j, n, z),
                    Bf.d4_besselj(n, z),
                    1e-9,
                )

        n = 0
        self.assertAlmostEqual(
            self._df(j, n, z),
            Bf.d1_besselj(n, z),
            1e-9,
        )
        self.assertAlmostEqual(
            self._d2f(j, n, z),
            Bf.d2_besselj(n, z),
            1e-9,
        )
        self.assertAlmostEqual(
            self._d3f(j, n, z),
            Bf.d3_besselj(n, z),
            1e-9,
        )
        self.assertAlmostEqual(
            self._d4f(j, n, z),
            Bf.d4_besselj(n, z),
            1e-9,
        )

    def test_bessely_and_derivatives(self):
        def y(n, z):
            return sqrt(pi / 2 / z) * yv(n + 0.5, z)

        for n in self.rng.integers(0, 10, size=10):
            for _ in np.arange(10):
                z = self.random_z()
                self.assertAlmostEqual(y(n, z), Bf.bessely(n, z), 1e-9)
                self.assertAlmostEqual(
                    self._df(y, n, z),
                    Bf.d1_bessely(n, z),
                    1e-9,
                )
                self.assertAlmostEqual(
                    self._d2f(y, n, z),
                    Bf.d2_bessely(n, z),
                    1e-9,
                )
                self.assertAlmostEqual(
                    self._d3f(y, n, z),
                    Bf.d3_bessely(n, z),
                    1e-9,
                )
                self.assertAlmostEqual(
                    self._d4f(y, n, z),
                    Bf.d4_bessely(n, z),
                    1e-9,
                )

        n = 0
        self.assertAlmostEqual(
            self._df(y, n, z),
            Bf.d1_bessely(n, z),
            1e-9,
        )
        self.assertAlmostEqual(
            self._d2f(y, n, z),
            Bf.d2_bessely(n, z),
            1e-9,
        )
        self.assertAlmostEqual(
            self._d3f(y, n, z),
            Bf.d3_bessely(n, z),
            1e-9,
        )
        self.assertAlmostEqual(
            self._d4f(y, n, z),
            Bf.d4_bessely(n, z),
            1e-9,
        )

    def test_hankelh1_and_derivatives(self):
        def h1(n, z):
            return sqrt(pi / z / 2) * hankel1(n + 0.5, z)

        for n in self.rng.integers(0, 10, size=10):
            for _ in np.arange(10):
                z = self.random_z()
                self.assertAlmostEqual(h1(n, z), Bf.hankelh1(n, z))
                self.assertAlmostEqual(
                    self._df(h1, n, z),
                    Bf.d1_hankelh1(n, z),
                    1e-9,
                )
                self.assertAlmostEqual(
                    self._d2f(h1, n, z),
                    Bf.d2_hankelh1(n, z),
                    1e-9,
                )
                self.assertAlmostEqual(
                    self._d3f(h1, n, z),
                    Bf.d3_hankelh1(n, z),
                    1e-9,
                )
                self.assertAlmostEqual(
                    self._d4f(h1, n, z),
                    Bf.d4_hankelh1(n, z),
                    1e-9,
                )
                self.assertAlmostEqual(
                    self._d5f(h1, n, z),
                    Bf.d5_hankelh1(n, z),
                    1e-9,
                )

        n = 0
        self.assertAlmostEqual(
            self._df(h1, n, z),
            Bf.d1_hankelh1(n, z),
            1e-9,
        )
        self.assertAlmostEqual(
            self._d2f(h1, n, z),
            Bf.d2_hankelh1(n, z),
            1e-9,
        )
        self.assertAlmostEqual(
            self._d3f(h1, n, z),
            Bf.d3_hankelh1(n, z),
            1e-9,
        )
        self.assertAlmostEqual(
            self._d4f(h1, n, z),
            Bf.d4_hankelh1(n, z),
            1e-9,
        )
        self.assertAlmostEqual(
            self._d5f(h1, n, z),
            Bf.d5_hankelh1(n, z),
            1e-9,
        )

    def test_hankelh2_and_derivatives(self):
        def h2(n, z):
            return sqrt(pi / z / 2) * hankel2(n + 0.5, z)

        for n in self.rng.integers(0, 10, size=10):
            for _ in np.arange(10):
                z = self.random_z()
                self.assertAlmostEqual(h2(n, z), Bf.hankelh2(n, z), 1e-9)
                self.assertAlmostEqual(
                    self._df(h2, n, z),
                    Bf.d1_hankelh2(n, z),
                    1e-9,
                )
                self.assertAlmostEqual(
                    self._d2f(h2, n, z),
                    Bf.d2_hankelh2(n, z),
                    1e-9,
                )
                self.assertAlmostEqual(
                    self._d3f(h2, n, z),
                    Bf.d3_hankelh2(n, z),
                    1e-9,
                )
                self.assertAlmostEqual(
                    self._d4f(h2, n, z),
                    Bf.d4_hankelh2(n, z),
                    1e-9,
                )
                self.assertAlmostEqual(
                    self._d5f(h2, n, z),
                    Bf.d5_hankelh2(n, z),
                    1e-9,
                )

        n = 0
        self.assertAlmostEqual(
            self._df(h2, n, z),
            Bf.d1_hankelh2(n, z),
            1e-9,
        )
        self.assertAlmostEqual(
            self._d2f(h2, n, z),
            Bf.d2_hankelh2(n, z),
            1e-9,
        )
        self.assertAlmostEqual(
            self._d3f(h2, n, z),
            Bf.d3_hankelh2(n, z),
            1e-9,
        )
        self.assertAlmostEqual(
            self._d4f(h2, n, z),
            Bf.d4_hankelh2(n, z),
            1e-9,
        )
        self.assertAlmostEqual(
            self._d5f(h2, n, z),
            Bf.d5_hankelh2(n, z),
            1e-9,
        )

    def test_adaptive_derivative(self):
        def j(n, z):
            return sqrt(pi / 2 / z) * jv(n + 0.5, z)

        def y(n, z):
            return sqrt(pi / 2 / z) * yv(n + 0.5, z)

        def h1(n, z):
            return sqrt(pi / z / 2) * hankel1(n + 0.5, z)

        def h2(n, z):
            return sqrt(pi / z / 2) * hankel2(n + 0.5, z)

        dfdzs = [self._f, self._df, self._d2f, self._d3f, self._d4f, self._d5f]

        for n in self.rng.integers(0, 10, size=10):
            for _ in np.arange(10):
                z = self.random_z()
                for i, df in enumerate(dfdzs):
                    self.assertAlmostEqual(
                        df(j, n, z),
                        Bf.adaptive_derivative_besselj(n, z, i),
                        1e-9,
                    )
                    self.assertAlmostEqual(
                        df(y, n, z),
                        Bf.adaptive_derivative_bessely(n, z, i),
                        1e-8,
                    )
                    self.assertAlmostEqual(
                        df(h1, n, z),
                        Bf.adaptive_derivative_hankelh1(n, z, i),
                        1e-8,
                    )
                    self.assertAlmostEqual(
                        df(h2, n, z),
                        Bf.adaptive_derivative_hankelh2(n, z, i),
                        1e-8,
                    )

        n = 0
        for i, df in enumerate(dfdzs):
            self.assertAlmostEqual(
                df(j, n, z),
                Bf.adaptive_derivative_besselj(n, z, i),
                1e-9,
            )
            self.assertAlmostEqual(
                df(y, n, z),
                Bf.adaptive_derivative_bessely(n, z, i),
                1e-9,
            )
            self.assertAlmostEqual(
                df(h1, n, z),
                Bf.adaptive_derivative_hankelh1(n, z, i),
                1e-9,
            )
            self.assertAlmostEqual(
                df(h2, n, z),
                Bf.adaptive_derivative_hankelh2(n, z, i),
                1e-9,
            )

    # -------------------------------------------------------------------------
    # Helper
    # -------------------------------------------------------------------------

    @staticmethod
    def _f(func, n, z):
        return func(n, z)

    @staticmethod
    def _df(func, n, z):
        out = -func(n + 1, z) + (n / z) * func(n, z)
        return out

    @classmethod
    def _d2f(cls, func, n, z):
        out = -cls._df(func, n + 1, z)
        out -= n / z**2 * func(n, z)
        out += n / z * cls._df(func, n, z)
        return out

    @classmethod
    def _d3f(cls, func, n, z):
        out = -cls._d2f(func, n + 1, z)
        out += 2 * n / z**3 * func(n, z)
        out -= n / z**2 * cls._df(func, n, z)
        out -= n / z**2 * cls._df(func, n, z)
        out += n / z * cls._d2f(func, n, z)
        return out

    @classmethod
    def _d4f(cls, func, n, z):
        out = -cls._d3f(func, n + 1, z)
        out -= 6 * n / z**4 * func(n, z)
        out += 3 * 2 * n / z**3 * cls._df(func, n, z)
        out -= 3 * n / z**2 * cls._d2f(func, n, z)
        out += n / z * cls._d3f(func, n, z)
        return out

    @classmethod
    def _d5f(cls, func, n, z):
        out = -cls._d4f(func, n + 1, z)
        out += 24 * n / z**5 * func(n, z)
        out -= 24 * n / z**4 * cls._df(func, n, z)
        out += 12 * n / z**3 * cls._d2f(func, n, z)
        out -= 4 * n / z**2 * cls._d3f(func, n, z)
        out += n / z * cls._d4f(func, n, z)
        return out


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np

from osaft.core.functions import exp
from osaft.core.functions import (
    spherical_2_cartesian_coordinates as s2c_coord,
)
from osaft.core.variable import ActiveVariable, PassiveVariable
from osaft.plotting.datacontainers.base_velocity_datacontainer import (
    VelocityFieldData,
)
from osaft.solutions.base_scattering import BaseScattering

NDArray = np.ndarray


class ScatteringFieldData(VelocityFieldData, ABC):
    """Container for plotting data for the scattering field

    :param sol: solution
    :param r_min: start point for radial plotting range
    :param r_max: end point for tangential plotting range
    :param res: resolution, if tuple (radial resolution, tangential resolution)
    :param n_quiver_points: number of points along z-axis for quiver plot
    :param symmetric: if ``True`` the dataset is assumed to be symmetric
    """

    def __init__(
        self,
        sol: BaseScattering,
        r_min: float,
        r_max: float,
        res: int | tuple[int, int] = 100,
        n_quiver_points: int | tuple[int, int] = 21,
        symmetric: bool = True,
    ) -> None:
        """Constructor method"""

        VelocityFieldData.__init__(
            self,
            sol,
            r_min,
            r_max,
            res,
            n_quiver_points,
            symmetric,
        )

        # Velocities for the scattering field (complex valued)
        self._u = ActiveVariable(self._reset_complex_array, "x coordinate")
        self._w = ActiveVariable(self._reset_complex_array, "z coordinate")

        self._u.is_computed_by(self._sol, self._grid, self._mode)
        self._w.is_computed_by(self._sol, self._grid, self._mode)

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def get_velocity_magnitude(
        self,
        instantaneous: bool = True,
        phase: float = 0,
        mode: None | int = None,
    ) -> tuple[NDArray, NDArray, NDArray]:
        """Returns the velocity magnitudes

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the magnitude of the velocity field at
        these points

        :param instantaneous: if `True` the instant magnitude is returned
        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        """
        self._mode.value = mode
        if instantaneous:
            return self.x, self.z, self._velocity_norm_inst(phase)
        else:
            return self.x, self.z, self.u_norm

    def get_displacement_magnitude(
        self,
        instantaneous: bool = True,
        phase: float = 0,
        mode: None | int = None,
    ) -> tuple[NDArray, NDArray, NDArray]:
        """Returns the displacement magnitudes

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the magnitude of the displacement field at
        these points

        :param instantaneous: if `True` the instant magnitude is returned
        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        """
        self._mode.value = mode
        if instantaneous:
            return self.x, self.z, self._disp_norm_inst(phase)
        else:
            return self.x, self.z, self.u_norm / self.sol.omega

    def get_velocity_vector(
        self,
        phase: float = 0,
        mode: None | int = None,
    ) -> tuple[NDArray, NDArray, NDArray, NDArray]:
        """Returns the velocity vectors

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the velocity field vectors at
        these points

        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        """
        self._mode.value = mode
        if phase == 0:
            return self.x, self.z, self.u.real, self.w.real
        else:
            u = self.u * np.exp(-1j * phase)
            w = self.w * np.exp(-1j * phase)
            return self.x, self.w, u.real, w.real

    def get_displacement_vector(
        self,
        phase: float = 0,
        mode: None | int = None,
    ) -> tuple[NDArray, NDArray, NDArray, NDArray]:
        """Returns the displacement vectors

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the displacement field vectors at
        these points

        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        """
        self._mode.value = mode
        if phase == 0:
            u = 1j * self.u / self.sol.omega
            w = 1j * self.w / self.sol.omega
        else:
            u = self.u * np.exp(-1j * phase) / self.sol.omega
            w = self.w * np.exp(-1j * phase) / self.sol.omega
        return self.x, self.w, u.real, w.real

    # -------------------------------------------------------------------------
    # Private Methods
    # -------------------------------------------------------------------------

    def _velocity_norm_inst(
        self,
        phase: float,
    ) -> NDArray:
        """Instantaneous velocity magnitude

        :param phase: phase angle
        """
        if phase == 0:
            return np.hypot(self.u.real, self.w.real)
        else:
            u = self.u * exp(-1j * phase)
            w = self.w * exp(-1j * phase)
            return np.hypot(u.real, w.real)

    def _disp_norm_inst(
        self,
        phase: float,
    ) -> NDArray:
        """Instantaneous displacement magnitude

        :param phase: phase angle
        """
        u = 1j * self.u * exp(-1j * phase) / self.sol.omega
        w = 1j * self.w * exp(-1j * phase) / self.sol.omega
        norm = np.hypot(u.real, w.real)
        return norm

    def _compute_data(self) -> None:
        """Compute the velocity data and store them in respective attributes

        Scattering velocity field is computed on the polar grid and transformed
        to Cartesian coordinates for plotting.
        """
        arr_r = self.grid.arr_r
        arr_theta = self.grid.arr_theta

        for index, r in enumerate(arr_r):
            # Indexing
            min_index = index * self.grid.res_theta
            max_index = min_index + self.grid.res_theta

            # Coordinates
            xs, zs = s2c_coord(r, arr_theta)
            self._x.value[min_index:max_index] = xs
            self._z.value[min_index:max_index] = zs

            # Displacements Velocities
            u, w = self._compute_velocities(r, arr_theta)
            self._u.value[min_index:max_index] = u
            self._w.value[min_index:max_index] = w
            self._v_norm.value[min_index:max_index] = np.hypot(
                np.abs(u),
                np.abs(w),
            )

    def _compute_data_quiver(self) -> None:
        """Compute the velocity data for the quiver and store them in
        respective attributes

        Scattering velocity field is computed on the cartesian grid
        """
        arr_r = self.grid_quiver.flat_mesh_r
        arr_theta = self.grid_quiver.flat_mesh_theta
        u = np.zeros_like(arr_r, dtype=np.complex_)
        w = np.zeros_like(arr_r, dtype=np.complex_)
        for i, (r, theta) in enumerate(zip(arr_r, arr_theta)):
            u[i], w[i] = self._compute_velocities(r, theta)

        # convert vectors from polar coordinates to cartesian
        u_cartesian = np.cos(arr_theta) * u - np.sin(arr_theta) * w
        w_cartesian = np.sin(arr_theta) * u + np.cos(arr_theta) * w

        self._z_quiver.value[:] = self.grid_quiver.flat_mesh_z
        self._x_quiver.value[:] = self.grid_quiver.flat_mesh_x
        self._u_quiver.value[:] = u_cartesian
        self._w_quiver.value[:] = w_cartesian

    def get_velocity_quiver(
        self,
        phase: float = 0,
        mode: None | int = None,
        scattered: bool = True,
        incident: bool = True,
        symmetric: bool = True,
        animation: bool = True,
    ) -> tuple[NDArray, NDArray, NDArray, NDArray, float]:
        """Returns the velocity potential

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the magnitude of the pressure field at
        these points

        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        :param scattered: option if scattering field is plotted
        :param incident: option if incident field is plotted
        :param symmetric: return symmetric solution
        :param animation: are the arrows used in an animation
        """
        self._mode.value = mode
        self.scattered = scattered
        self.incident = incident

        u = self.u_quiver * exp(-1j * phase)
        w = self.w_quiver * exp(-1j * phase)

        # remove nan from u and w
        bools = np.isnan(u)
        u[bools] = 0
        w[bools] = 0

        if animation:
            norm = np.max(np.hypot(abs(u), abs(w)))
        else:
            norm = np.max(np.hypot(u.real, w.real))
        scale = self.n_quiver_points / 2 * norm

        z = self.z_quiver
        x = self.x_quiver
        if symmetric:
            z = np.concatenate([z, z])
            u = np.concatenate([u, u])

            x = np.concatenate([-x, x])
            w = np.concatenate([-w, w])

        return z, x, u.real, w.real, scale

    @abstractmethod
    def _compute_velocities(
        self,
        r: float,
        theta: float | NDArray,
    ) -> tuple[NDArray, NDArray]:
        """Compute the velocity data and return it

        :param arr_r: array of r-coordinates
        :param theta: polar angle
        """
        pass  # pragma: no cover


# -----------------------------------------------------------------------------
# ParticleScatteringData
# -----------------------------------------------------------------------------


class ParticleScatteringData(ScatteringFieldData):
    """Container for plotting data for the displacement field in the particle

    :param sol: solution
    :param res: resolution, if tuple (radial resolution, tangential resolution)
    :param n_quiver_points: anchor points along z for quiver arrows
    """

    def __init__(
        self,
        sol: BaseScattering,
        res: int | tuple[int, int] = 100,
        n_quiver_points: int = 21,
    ) -> None:
        """Constructor method"""

        r_min = self._EPS_r_min
        r_max = sol.R_0
        super().__init__(
            sol,
            r_min,
            r_max,
            res,
            n_quiver_points,
        )

    # -------------------------------------------------------------------------
    # Private Methods
    # -------------------------------------------------------------------------

    def _compute_velocities(
        self,
        r: NDArray,
        theta: float,
    ):
        """Compute the velocity data and return it

        :param arr_r: array of r-coordinates
        :param theta: polar angle
        """
        u = self.sol.radial_particle_velocity(r, theta, 0, self.mode)
        w = self.sol.tangential_particle_velocity(r, theta, 0, self.mode)
        return u, w


# -----------------------------------------------------------------------------
# FluidScatteringData
# -----------------------------------------------------------------------------


class FluidScatteringData(ScatteringFieldData):
    """Container for plotting data for the scattering field in the fluid

    :param sol: solution
    :param r_max: end point for tangential plotting range
    :param res: resolution, if tuple (radial resolution, tangential resolution)
    :param n_quiver_points: anchor points along z for quiver arrows
    """

    def __init__(
        self,
        sol: BaseScattering,
        r_max: float,
        scattering: bool = True,
        incident: bool = True,
        res: int | tuple[int, int] = 100,
        n_quiver_points: int = 21,
    ) -> None:
        """Constructor method"""

        r_min = sol.R_0
        super().__init__(
            sol,
            r_min,
            r_max,
            res,
            n_quiver_points,
        )

        self._scattered = PassiveVariable(
            scattering,
            "option: plot scattering field",
        )
        self._incident = PassiveVariable(
            incident,
            "option: plot incident field",
        )

        self._vel_pot = ActiveVariable(
            self._reset_complex_array,
            "velocity potential",
        )
        self._pressure = ActiveVariable(
            self._reset_complex_array,
            "acoustic pressure",
        )

        # Set links
        self._u.is_computed_by(self._scattered, self._incident)
        self._w.is_computed_by(self._scattered, self._incident)
        self._vel_pot.is_computed_by(self._sol, self._grid, self._mode)
        self._pressure.is_computed_by(self._sol, self._grid, self._mode)

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def get_pressure(
        self,
        instantaneous: bool = True,
        phase: float = 0,
        mode: None | int = None,
        scattered: bool = True,
        incident: bool = True,
    ) -> tuple[NDArray, NDArray, NDArray]:
        """Returns the velocity potential

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the magnitude of the pressure field at
        these points

        :param instantaneous: if ``True``, then instantaneous field
        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        :param scattered: option if scattering field is plotted
        :param incident: option if incident field is plotted
        """
        self._mode.value = mode
        self.scattered = scattered
        self.incident = incident
        if instantaneous:
            pressure_inst = self.pressure * exp(-1j * phase)
            return self.x, self.z, pressure_inst.real
        else:
            return self.x, self.z, np.abs(self.pressure)

    def get_velocity_potential(
        self,
        instantaneous: bool = True,
        phase: float = 0,
        mode: None | int = None,
        scattered: bool = True,
        incident: bool = True,
    ) -> tuple[NDArray, NDArray, NDArray]:
        """Returns the velocity potential

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the magnitude of the velocity potential field at
        these points

        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        :param scattered: option if scattering field is plotted
        :param incident: option if incident field is plotted
        """
        self._mode.value = mode
        self.scattered = scattered
        self.incident = incident
        if instantaneous:
            vel_pot_inst = self.velocity_potential * exp(-1j * phase)
            return self.x, self.z, vel_pot_inst.real
        else:
            return self.x, self.z, np.abs(self.velocity_potential)

    def get_velocity_magnitude(
        self,
        instantaneous: bool = True,
        phase: float = 0,
        mode: None | int = None,
        scattered: bool = True,
        incident: bool = True,
    ) -> tuple[NDArray, NDArray, NDArray]:
        """Returns the velocity magnitudes

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the magnitude of the velocity field at
        these points

        :param instantaneous: option if instantaneous or average magnitude
        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        :param scattered: option if scattering field is plotted
        :param incident: option if incident field is plotted
        """
        self.scattered = scattered
        self.incident = incident
        return super().get_velocity_magnitude(instantaneous, phase, mode)

    def get_velocity_vector(
        self,
        phase: float = 0,
        mode: None | int = None,
        scattered: bool = True,
        incident: bool = True,
    ) -> tuple[NDArray, NDArray, NDArray, NDArray]:
        """Returns the velocity vectors

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the velocity field vectors at
        these points

        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        :param scattered: option if scattering field is plotted
        :param incident: option if incident field is plotted
        """
        self.scattered = scattered
        self.incident = incident
        return super().get_velocity_vector(phase, mode)

    def get_displacement_vector(
        self,
        phase: float = 0,
        mode: None | int = None,
        scattered: bool = True,
        incident: bool = True,
    ) -> tuple[NDArray, NDArray, NDArray, NDArray]:
        """Returns the displacement vectors

        returns the coordinates of the points of the plotting grid `x`,
        `y` and the displacement field vectors at
        these points

        :param phase: phase angle
        :param mode: mode of oscillation, if `None` superposition
        :param scattered: option if scattering field is plotted
        :param incident: option if incident field is plotted
        """
        self.scattered = scattered
        self.incident = incident
        return super().get_displacement_vector(phase, mode)

    # -------------------------------------------------------------------------
    # Getters for Dependent Variables
    # -------------------------------------------------------------------------
    @property
    def velocity_potential(self) -> np.array:
        if np.all(self._vel_pot.value == 0):
            self._compute_velocity_potential()
        return self._vel_pot.value

    @property
    def pressure(self) -> np.array:
        if np.all(self._pressure.value == 0):
            self._compute_pressure()
        return self._pressure.value

    @property
    def scattered(self) -> bool:
        """Option if scattering field is plotted

        :getter: returns True (plot field) or False (do not plot field)
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._scattered.value

    @scattered.setter
    def scattered(self, value: bool) -> None:
        self._scattered.value = value

    @property
    def incident(self) -> bool:
        """Option if incident field is plotted

        :getter: returns True (plot field) or False (do not plot field)
        :setter: automatically invokes
            :meth:`osaft.core.variable.BaseVariable.notify`
        """
        return self._incident.value

    @incident.setter
    def incident(self, value: bool) -> None:
        self._incident.value = value

    def _compute_pressure(self) -> None:
        """Compute the velocity data and store them in respective attributes

        Scattering velocity field is computed on the polar grid and transformed
        to Cartesian coordinates for plotting.
        """
        arr_r = self.grid.arr_r
        arr_theta = self.grid.arr_theta

        for index, r in enumerate(arr_r):
            # Indexing
            min_index = index * self.grid.res_theta
            max_index = min_index + self.grid.res_theta

            # Coordinates
            xs, zs = s2c_coord(r, arr_theta)
            self._x.value[min_index:max_index] = xs
            self._z.value[min_index:max_index] = zs

            # Pressure
            pressure = self.sol.pressure(
                r,
                arr_theta,
                0,
                self.scattered,
                self.incident,
                self.mode,
            )
            self._pressure.value[min_index:max_index] = pressure

    def _compute_velocity_potential(self) -> None:
        """Compute the velocity data and store them in respective attributes

        Scattering velocity field is computed on the polar grid and transformed
        to Cartesian coordinates for plotting.
        """
        arr_r = self.grid.arr_r
        arr_theta = self.grid.arr_theta

        for index, r in enumerate(arr_r):
            # Indexing
            min_index = index * self.grid.res_theta
            max_index = min_index + self.grid.res_theta

            # Coordinates
            xs, zs = s2c_coord(r, arr_theta)
            self._x.value[min_index:max_index] = xs
            self._z.value[min_index:max_index] = zs

            # Velocity Potential
            vel_pot = self.sol.velocity_potential(
                r,
                arr_theta,
                0,
                self.scattered,
                self.incident,
                self.mode,
            )
            self._vel_pot.value[min_index:max_index] = vel_pot

    # -------------------------------------------------------------------------
    # Private Methods
    # -------------------------------------------------------------------------

    def _compute_velocities(
        self,
        r: float,
        theta: float | NDArray,
    ) -> tuple[NDArray, NDArray]:
        """Compute the velocity data and return it

        :param r: radial coordinate
        :param theta: single value or array of tangential coordinates
        """
        u = self.sol.radial_acoustic_fluid_velocity(
            r,
            theta,
            0,
            self.scattered,
            self.incident,
            self.mode,
        )
        w = self.sol.tangential_acoustic_fluid_velocity(
            r,
            theta,
            0,
            self.scattered,
            self.incident,
            self.mode,
        )
        return u, w


if __name__ == "__main__":
    pass



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import FuncAnimation
from matplotlib.quiver import Quiver

from osaft.core.functions import pi
from osaft.plotting.datacontainers.scattering_datacontainer import (
    ParticleScatteringData,
)
from osaft.plotting.datacontainers.wireframe_datacontainer import (
    ParticleWireframeData,
)
from osaft.plotting.scattering.base_plotter import BaseScatteringPlotter
from osaft.plotting.scattering.base_scattering_plots import (
    BaseScatteringPlots,
)
from osaft.plotting.scattering.tri_plotter import TriangulationPlotter
from osaft.plotting.scattering.wireframe_plotter import WireframePlotter
from osaft.solutions.base_scattering import BaseScattering

NDArray = np.ndarray


class ParticleScatteringPlot(BaseScatteringPlots):
    """Class for plotting scattering field of the particle

    Plots the acoustic field inside the particle using Matplotlib
    tricontourf or tripcolor plotting methods.

    :param sol: solution to be plotted
    :param resolution: if tuple (radial resolution, tangential resolution)
    :param n_quiver_points: anchor points along z for quiver
    :param cmap: color map
    """

    def __init__(
        self,
        sol: BaseScattering,
        resolution: int | tuple[int, int] = 100,
        n_quiver_points: int = 21,
        cmap: str = "winter",
    ):
        """Constructor method"""
        self.data = ParticleScatteringData(
            sol,
            resolution,
            n_quiver_points,
        )
        self.plotter = TriangulationPlotter(True, cmap)

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def plot_velocity(
        self,
        inst: bool = True,
        phase: float = 0,
        mode: None | int = None,
        displacement: bool = False,
        symmetric: bool = True,
        quiver_color: None | str = None,
        tripcolor: bool = False,
        ax: None | plt.Axes = None,
        **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Tricontourf plot for acoustic velocity field of the particle

        Plots the velocity amplitude of the first-order acoustic velocity field
        of the particle using Matplotlib's
        `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tricontour.html>`_
        or
        `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tripcolor.html>`_
        if ``tripcolor = True``.

        :param inst: if ``True`` instantaneous amplitude is plotted
        :param phase: phase :math:`[0, 2\\pi]`
        :param mode: mode of oscillation
        :param displacement: if ``True`` displacement, else velocity plot
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param quiver_color: color of the quiver arrows, if None: no arrows
        :param tripcolor: switches between tripcolor and tricontourf plot
        :param ax: Axes object
        :param kwargs: passed through to tricontourf()
        """

        fig, ax = self._triangulation_plot(
            tripcolor=tripcolor,
            inst=inst,
            phase=phase,
            mode=mode,
            displacement=displacement,
            symmetric=symmetric,
            ax=ax,
            **kwargs,
        )

        if quiver_color:
            # particle plots does not a distinction between scattered and
            # incident field
            scattered = incident = False
            self._overlay_quiver(
                ax,
                phase,
                scattered,
                incident,
                mode,
                symmetric,
                quiver_color,
                animation=False,
            )

        return fig, ax

    def animate_velocity(
        self,
        frames: int = 64,
        interval: float = 100.0,
        mode: None | int = None,
        displacement: bool = True,
        symmetric: bool = True,
        quiver_color: None | str = None,
        tripcolor: bool = False,
        ax: None | plt.Axes = None,
        **kwargs,
    ) -> FuncAnimation:
        """Tricontourf animation for acoustic velocity field of the particle

        Animates the velocity amplitude of the first-order acoustic velocity
        field of the particle over one period using Matplotlib's
        `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tricontour.html>`_
        or
        `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tripcolor.html>`_
        if ``tripcolor = True``.


        :param frames: number of frames for the animation
        :param interval: interval between frames in ms
        :param mode: mode of oscillation
        :param displacement: if ``True`` displacement, else velocity plot
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param quiver_color: color of the quiver arrows, if None: no arrows
        :param tripcolor: switches between tripcolor and tricontourf plot
        :param ax: Axes object
        :param kwargs: passed through to tricontourf()
        """

        if quiver_color:

            fig, ax = self.plot_velocity(
                inst=False,
                mode=mode,
                symmetric=symmetric,
            )

            # no scattered or incident field for scatterer
            scattered = incident = False
            Q = self._overlay_quiver(
                ax,
                0,
                scattered,
                incident,
                mode,
                symmetric,
                animation=True,
                quiver_color=quiver_color,
            )
            annotation = self.plotter._annotate_phase(ax, 0)

            def animate_quiver(frame: int, frames: int, Q: Quiver, annotation):
                phase = frame / frames * 2 * pi

                _, _, u, w, _ = self.data.get_velocity_quiver(
                    phase,
                    mode,
                    scattered,
                    incident,
                    symmetric,
                    animation=True,
                )

                self.plotter._update_phase_annotation(annotation, phase)

                Q.set_UVC(u, w)

                return (Q,)

            anim = FuncAnimation(
                fig,
                animate_quiver,
                frames=frames,
                interval=interval,
                blit=False,
                fargs=(frames, Q, annotation),
            )

            return anim
        else:
            return self._triangulation_animation(
                tripcolor=tripcolor,
                frames=frames,
                interval=interval,
                mode=mode,
                displacement=displacement,
                symmetric=symmetric,
                ax=ax,
                **kwargs,
            )

    # -------------------------------------------------------------------------
    # Private Methods
    # -------------------------------------------------------------------------

    def _triangulation_plot(
        self,
        tripcolor: bool,
        inst: bool,
        phase: float,
        mode: None | int,
        displacement: bool,
        symmetric: bool,
        ax: None | plt.Axes,
        **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Helper function for tripcolor/tricontourf plot

        :param tripcolor: if ``True`` tripcolor, else tricontourf plot
        :param inst: if ``True`` instantaneous amplitude is plotted
        :param phase: phase :math:`[0, 2\\pi]`
        :param mode: mode of oscillation
        :param displacement: if ``True`` displacement, else velocity plot
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param ax: Axes object
        :param kwargs: passed through to plotting method
        """

        # Data
        if displacement:
            X, Z, C = self.data.get_displacement_magnitude(inst, phase, mode)
            cbar_label = "Displacement [m]"
        else:
            X, Z, C = self.data.get_velocity_magnitude(inst, phase, mode)
            cbar_label = "Velocity [m/s]"
        # Replace nan in the center
        C[0] = C[1]
        # Plot
        fig, ax, _, _, _ = self.plotter.plot(
            X=Z,
            Y=X,
            C=C,
            radius=self.data.sol.R_0,
            symmetric=symmetric,
            tripcolor=tripcolor,
            cbar_label=cbar_label,
            use_diverging_cmap=False,
            ax=ax,
            **kwargs,
        )
        return fig, ax

    def _triangulation_animation(
        self,
        tripcolor: bool,
        frames: int,
        interval: float,
        mode: None | int,
        displacement: bool,
        symmetric: bool,
        ax: None | plt.Axes,
        **kwargs,
    ) -> FuncAnimation:
        """Helper function for tripcolor/tricontourf animation

        :param tripcolor: if ``True`` tripcolor, else tricontourf plot
        :param frames: number of frames for the animation
        :param interval: interval between frames in ms
        :param mode: mode of oscillation
        :param displacement: if ``True`` displacement, else velocity plot
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param ax: Axes object
        :param kwargs: passed through to tricontourf()
        """

        # Data function for animation
        def data_func(phase: float) -> tuple[NDArray, NDArray, NDArray]:
            """Returns the velocity field for given phase
            Closure is used to fix all additional parameters
            :param phase: phase
            """
            if displacement:
                return self.data.get_displacement_magnitude(True, phase, mode)
            else:
                return self.data.get_velocity_magnitude(True, phase, mode)

        # Color bar label
        if displacement:
            cbar_label = "Displacement [m]"
        else:
            cbar_label = "Velocity [m/s]"

        # Get norm
        if displacement:
            _, _, C = self.data.get_displacement_magnitude(False, mode=mode)
        else:
            _, _, C = self.data.get_velocity_magnitude(False, mode=mode)

        return self.plotter.animate(
            frames=frames,
            interval=interval,
            tripcolor=tripcolor,
            symmetric=symmetric,
            cbar_label=cbar_label,
            use_diverging_cmap=False,
            animate_meth=data_func,
            C_norm=C,
            radius=self.data.sol.R_0,
            ax=ax,
            **kwargs,
        )

    def plot_velocity_evolution(
        self,
        inst: bool = True,
        mode: None | int = None,
        displacement: bool = False,
        symmetric: bool = True,
        quiver_color: str = None,
        tripcolor: bool = False,
        layout: tuple[int, int] = (3, 3),
        **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Tricontourf plot for acoustic velocity field evolution of the
        particle

        Plots the velocity amplitude of the first-order acoustic velocity
        field of the fluid over one period at different phases using
        Matplotlib's
        `tricontourf
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tricontour.html>`_
        or
        `tripcolor
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tripcolor.html>`_
        if ``tripcolor = True``.

        The first phase value is always :math:`0\\pi` and the last one
        :math:`2\\pi`. The total number of plots and, hence, also the steps
        between the different phase values is the defined by the product of the
        ``layout`` tuple.

        :param inst: if ``True`` instantaneous amplitude is plotted
        :param mode: mode of oscillation
        :param displacement: if ``True`` displacement, else velocity plot
        :param symmetric: if ``True`` the symmetry of the solution is used
        :param quiver_color: color of the quiver arrows, if None: no arrows
        :param tripcolor: switches between tripcolor and tricontourf plot
        :param layout: number of rows and columns for plotting
        :param kwargs: passed through to the parent subplots command
        """

        n_row, n_col = layout
        n = n_col * n_row

        phases = np.linspace(0, 2, num=n)

        fig, axes = plt.subplots(
            n_row,
            n_col,
            sharex=True,
            sharey=True,
            **kwargs,
        )

        # Data
        if displacement:
            cbar_label = "Displacement [m]"
        else:
            cbar_label = "Velocity [m/s]"

        # Data function for subplots
        def data_func(phase: float) -> tuple[NDArray, NDArray, NDArray]:
            """Returns the velocity field for given phase
            Closure is used to fix all additional parameters
            :param phase: phase
            """
            if displacement:
                return self.data.get_displacement_magnitude(inst, phase, mode)
            else:
                return self.data.get_velocity_magnitude(inst, phase, mode)

        # Get norm
        if displacement:
            _, _, C_norm = self.data.get_displacement_magnitude(
                False,
                mode=mode,
            )
        else:
            _, _, C_norm = self.data.get_velocity_magnitude(False, mode=mode)

        for i, phase in enumerate(phases):

            row = i // n_col
            col = i % n_col
            ax = axes.flat[i]

            X, Z, C = data_func(phase=phase * np.pi)
            # Replace nan in the center
            C[0] = C[1]

            # Plot
            _, _, cnf, cbar, _ = self.plotter.plot(
                X=Z,
                Y=X,
                C=C,
                radius=self.data.sol.R_0,
                symmetric=symmetric,
                tripcolor=tripcolor,
                cbar_label=cbar_label,
                use_diverging_cmap=False,
                ax=ax,
                vmin=0,
                vmax=1.01 * np.nanmax(C_norm),
            )
            if quiver_color:
                # particle plots does not a distinction between scattered and
                # incident field
                scattered = incident = False
                self._overlay_quiver(
                    ax,
                    phase,
                    scattered,
                    incident,
                    mode,
                    symmetric,
                    animation=False,
                    quiver_color=quiver_color,
                )
            # remove colorbar
            cbar.remove()

            ax.set_title(f"{phase:.2f}" + r"$\pi$")

            if row != (n_row - 1):
                ax.set_xlabel("")
            if col > 0:
                ax.set_ylabel("")

        fig.tight_layout()
        cbar = fig.colorbar(cnf, ax=axes.ravel().tolist())
        cbar.ax.set_ylabel(cbar_label)

        return fig, axes


class ParticleWireframePlot(BaseScatteringPlotter):
    """Plotting class for wireframe plot of the particle

    :param sol: solution to be plotted
    :param nbr_r_levels: number of circles shown in the wireframe
    :param nbr_theta_levels: number of radii shown in the wireframe
    :param resolution: resolution, if tuple `(radial res, tangential res)`
    :param scale_factor: scaling factor for the displacements, if ``None`` auto
    """

    def __init__(
        self,
        sol: BaseScattering,
        nbr_r_levels: int = 10,
        nbr_theta_levels: int = 10,
        resolution: int | tuple[int, int] = (100, 100),
        scale_factor: float = 0.1,
    ) -> None:
        # Call to parent class
        super().__init__(True)

        self.data = ParticleWireframeData(
            sol,
            nbr_r_levels,
            nbr_theta_levels,
            resolution,
            scale_factor,
        )
        self.plotter = WireframePlotter()

    # -------------------------------------------------------------------------
    # API
    # -------------------------------------------------------------------------

    def plot(
        self,
        phase: float = 0,
        mode: None | int = None,
        ax: None | plt.Axes = None,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Plot wireframe

        :param phase: phase to be plotted
        :param mode: mode to be plotted, if ``None`` superposition of all modes
        :param ax: if ``ax`` is passed, plot will be drawn on ``ax``
        """
        self.data.mode = mode
        radii, circles = self.data.get_displacements()
        fig, ax = self.plotter.plot(radii, circles, phase, ax)

        return fig, ax

    def animate(
        self,
        frames: int = 64,
        interval: float = 100.0,
        mode: None | int = None,
        ax: None | plt.Axes = None,
    ):
        """Animate wireframe

        :param frames: number of frames for the animation
        :param interval: interval between frames in milliseconds
        :param mode: mode to be plotted, if ``None`` superposition of all modes
        :param ax: if ``ax`` is passed, plot will be drawn on ``ax``
        """
        self.data.mode = mode
        anim = self.plotter.animate(
            frames=frames,
            interval=interval,
            deformed_radii=self.data.deformed_radii,
            deformed_circles=self.data.deformed_circles,
            ax=ax,
        )
        return anim

    def plot_evolution(
        self,
        mode: None | int = None,
        layout: tuple[int, int] = (3, 3),
        **kwargs,
    ) -> tuple[plt.Figure, plt.Axes]:
        """Plot wireframe over one period

        :param layout: subplot layout of the figure
        :param mode: mode to be plotted, if ``None`` superposition of all modes
        """

        self.data.mode = mode
        n_row, n_col = layout
        n = n_col * n_row

        phases = np.linspace(0, 2, num=n)

        fig, axes = plt.subplots(
            n_row,
            n_col,
            sharex=True,
            sharey=True,
            **kwargs,
        )

        for i, phase in enumerate(phases):

            row = i // n_col
            col = i % n_col
            ax = axes.flat[i]

            radii, circles = self.data.get_displacements()
            _, _ = self.plotter.plot(radii, circles, phase * np.pi, ax)

            ax.set_title(f"{phase:.2f}" + r"$\pi$")

            if row != (n_row - 1):
                ax.set_xlabel("")
            if col > 0:
                ax.set_ylabel("")

        fig.tight_layout()
        return fig, axes


if __name__ == "__main__":
    pass
